/* ==================================================================
 * Spread Client Extension for PHP
 * Copyright (c) 2001 George Schlossnagle
 * All rights reserved.
 * ==================================================================
 * This source code is made available free and without charge subject
 * to the terms of the PHP Group License as detailed in bundled LICENSE file.
 * ==================================================================
 * George Schlossnagle <george@omniti.com>
 * ==================================================================
*/

#include "config.h"

#include "php_spread.h"
#include "php_globals.h"
#include "php.h"
#include "ext/standard/php_string.h"
#include <sp.h>
#include <sys/time.h>

/* for function declarations */

/* This comes from php install tree */
#include "ext/standard/info.h"

/* declarations of functions to be exported */
PHP_FUNCTION(spread_connect);
PHP_FUNCTION(spread_multicast);
PHP_FUNCTION(spread_close);
PHP_FUNCTION(spread_error);
PHP_FUNCTION(spread_join);
PHP_FUNCTION(spread_receive);

static int le_conn;

/* list of exported functions */
static unsigned char a2_arg_force_ref[] = { 2, BYREF_NONE, BYREF_FORCE };

function_entry spread_functions[] = {
	PHP_FE(spread_connect, NULL)
	PHP_FE(spread_multicast, NULL)
	PHP_FE(spread_error, NULL)
	PHP_FE(spread_close, NULL)
	PHP_FE(spread_join, NULL)
	PHP_FE(spread_receive, NULL)
	{NULL, NULL, NULL}
};


/* module entry */
zend_module_entry spread_module_entry = {
    STANDARD_MODULE_HEADER,
	"spread",
	spread_functions,
	ZEND_MODULE_STARTUP_N(spread),
	PHP_MSHUTDOWN(spread),
	PHP_RINIT(spread),
	PHP_RSHUTDOWN(spread),
	PHP_MINFO(spread),
        NO_VERSION_YET,
	STANDARD_MODULE_PROPERTIES
};

zend_spread_globals spread_globals;

#ifdef COMPILE_DL_PHP_SPREAD
ZEND_GET_MODULE(spread)
#endif

/* UTILITIES */
	
static void _close_spread_pconn(zend_rsrc_list_entry *rsrc) {
	mailbox *mbox = (int *)rsrc->ptr;
    SP_disconnect(*mbox);
	free(mbox);
}


/* initialization file support */


ZEND_MODULE_STARTUP_D(spread)
{
    le_conn = zend_register_list_destructors_ex(NULL, _close_spread_pconn, "spread link persistent", module_number);
    REGISTER_LONG_CONSTANT("SP_LOW_PRIORITY", LOW_PRIORITY, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_MEDIUM_PRIORITY", MEDIUM_PRIORITY, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_HIGH_PRIORITY", HIGH_PRIORITY, CONST_CS | CONST_PERSISTENT);

    REGISTER_LONG_CONSTANT("SP_UNRELIABLE_MESS", UNRELIABLE_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_RELIABLE_MESS", RELIABLE_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_FIFO_MESS", FIFO_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_CAUSAL_MESS", CAUSAL_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_AGREED_MESS", AGREED_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_SAFE_MESS", SAFE_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("SP_REGULAR_MESS", REGULAR_MESS, CONST_CS | CONST_PERSISTENT);

    REGISTER_LONG_CONSTANT("REG_MEMB_MESS", REG_MEMB_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("TRANSITION_MESS", TRANSITION_MESS, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("CAUSED_BY_JOIN", CAUSED_BY_JOIN, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("CAUSED_BY_LEAVE", CAUSED_BY_LEAVE, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("CAUSED_BY_DISCONNECT", CAUSED_BY_DISCONNECT, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("CAUSED_BY_NETWORK", CAUSED_BY_NETWORK, CONST_CS | CONST_PERSISTENT);
    REGISTER_LONG_CONSTANT("MEMBERSHIP_MESS", MEMBERSHIP_MESS, CONST_CS | CONST_PERSISTENT);

	return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(spread)
{
	return SUCCESS;
}

PHP_RINIT_FUNCTION(spread)
{
	SpG(default_conn) = -1;
	return SUCCESS;
}

PHP_RSHUTDOWN_FUNCTION(spread)
{
	return SUCCESS;
}

PHP_MINFO_FUNCTION(spread)
{
}

PHP_GINIT_FUNCTION(spread)
{
	return SUCCESS;
}

PHP_GSHUTDOWN_FUNCTION(spread)
{
	return SUCCESS;
}


/* exported function definitions */

/* {{{ proto int spread_connect(string spreaddaemon [, string private_name])
  Open a persistent connection to a spread daemon.  Returns resource id# */
PHP_FUNCTION(spread_connect)
{
    char *spread_name;
    char private_name[128];
    mailbox *mbox;
    char private_group[MAX_GROUP_NAME];
    char *hashed_details;
    int hashed_details_length;
    zval **z_spread_name = NULL, **z_private_name = NULL;
    list_entry *le;

    switch(ZEND_NUM_ARGS()) {
        case 1: {
                if (zend_get_parameters_ex(1, &z_spread_name) == FAILURE) {
                    // return failure
                }
            }
            break;
        case 2: {
                if(zend_get_parameters_ex(2, &z_spread_name, &z_private_name)
                    == FAILURE) {
                    // return false
                }
            }
            break;
        default:
            WRONG_PARAM_COUNT;
            break;
    }
	SEPARATE_ZVAL(z_spread_name);
    convert_to_string_ex(z_spread_name);
    spread_name = (*z_spread_name)->value.str.val;
    if(z_private_name) {
        convert_to_string_ex(z_private_name);
        snprintf(private_name,MAX_PRIVATE_NAME,"%s",(*z_private_name)->value.str.val)
;
    }
    else {
        snprintf(private_name,MAX_PRIVATE_NAME,"php-%05d", getpid());
    }
    hashed_details_length = sizeof("spread__") + strlen(spread_name) +
        strlen(private_name);
    hashed_details = (char *) emalloc(hashed_details_length);
    sprintf(hashed_details, "spread_%s_%s", spread_name, private_name);
    /* look up spread connection in persistent_list */
    if (zend_hash_find(&EG(persistent_list), hashed_details,
        hashed_details_length, (void **) &le) == FAILURE)
    {
        list_entry new_le;
        int retval;

        mbox = (mailbox *) malloc(sizeof(int));
        if ((retval = SP_connect(spread_name, private_name, 0, 0, mbox,
            private_group)) != ACCEPT_SESSION)
        {	
			zend_error(E_WARNING, "Failed to connect to spread daemon %s, error returned was: %d", spread_name, retval);
           efree(hashed_details);
		   RETURN_FALSE;
        }
		else {
		}
        new_le.type = le_conn;
        new_le.ptr = mbox;
        if (zend_hash_update(&EG(persistent_list), hashed_details,
            hashed_details_length, (void *) &new_le, sizeof(list_entry),
            NULL) == FAILURE)
        {
            SP_disconnect(*mbox);
			free(mbox);
          	efree(hashed_details);
            RETURN_FALSE;
        }
    }
    else { /* we have a pre-existing connection */
        if (le->type != le_conn) {
            // return badly
			efree(hashed_details);
			RETURN_FALSE;
        }
        mbox = (mailbox *) le->ptr;
    }
    ZEND_REGISTER_RESOURCE(return_value, mbox, le_conn);
    efree(hashed_details);
	if(SpG(default_conn) != -1) {
		zend_list_delete(SpG(default_conn));
	}
	SpG(default_conn) = return_value->value.lval;
	zend_list_addref(return_value->value.lval); 
	return;
}
/* }}} */

/* {{{ proto int spread_multicast( string groupname | array groupnames, string message [, string resource id])
	Multicast a message to spread group(s), returns bytes sent */
PHP_FUNCTION(spread_multicast) {
    zval **group = NULL;
    zval **message = NULL;
    zval **optional_mbox = NULL;
    zval **mess_type_zval = NULL;
    zval **service_type_zval = NULL;
    int *mbox, default_mbox, service_type, mess_type;
    int sperrno;
    switch(ZEND_NUM_ARGS()) {
        case 2:
            if(zend_get_parameters_ex(2, &group, &message) == FAILURE) {
                RETURN_FALSE;
            }
            if(SpG(default_conn)) {
                default_mbox = SpG(default_conn);
            }
            else {
                RETURN_FALSE;
            }
            service_type = RELIABLE_MESS;
            mess_type = 1;
            break;
        case 3:
            if(zend_get_parameters_ex(3, &optional_mbox, &group, &message) == FAILURE) {
                RETURN_FALSE;
            }
            service_type = RELIABLE_MESS;
            mess_type = 1;
            break;
        case 5:
            if(zend_get_parameters_ex(4, &optional_mbox, 
                                      &service_type_zval, 
                                      &group, 
                                      &mess_type_zval, 
                                      &message) == FAILURE) 
            {
                RETURN_FALSE;
            }
            service_type = Z_LVAL_PP(service_type_zval);
            mess_type = Z_LVAL_PP(mess_type_zval);
            break;
        default:
            WRONG_PARAM_COUNT;
            break;
    }
	SEPARATE_ZVAL(group)
	SEPARATE_ZVAL(message)
	fprintf(stderr, "Calling Z_F_R2(mbox, optional_inbox,%d,...)", default_mbox);
    ZEND_FETCH_RESOURCE(mbox, int *, optional_mbox, default_mbox, "Spread-FD", le_conn);
    convert_to_string_ex(message);
	if((*group)->type == IS_STRING) {
		char *tmpgrp;
    	convert_to_string_ex(group);
		tmpgrp = estrndup((*group)->value.str.val,(*group)->value.str.len);
    	if (sperrno = (SP_multicast(*mbox, RELIABLE_MESS, 
			tmpgrp, 1, (*message)->value.str.len,
        	(*message)->value.str.val)) <0)
    	{
			zend_error(E_WARNING, "SP_mulicast error(%d)", sperrno);
			RETURN_FALSE;
    	}
		return_value->type = IS_LONG;
		return_value->value.lval = sperrno;
		return;
	}
	if((*group)->type == IS_ARRAY) {
		char groupnames[100][MAX_GROUP_NAME];
		HashPosition pos;
		zval *tmparr, **tmp;
		int n = 0;
		tmparr = *group;

		zend_hash_internal_pointer_reset_ex(tmparr->value.ht, &pos);
		while(zend_hash_get_current_data_ex(tmparr->value.ht, (void **) &tmp, &pos) == SUCCESS && n < 100) {
			convert_to_string_ex(tmp);
			strncpy(groupnames[n], (*tmp)->value.str.val, MAX_GROUP_NAME);
			n++;
			zend_hash_move_forward_ex(tmparr->value.ht, &pos);
		}
		if((sperrno = SP_multigroup_multicast(*mbox, service_type,
			n, (const char (*)[MAX_GROUP_NAME]) groupnames, mess_type, 
			(*message)->value.str.len, (*message)->value.str.val)) <0)
		{
            zend_error(E_WARNING, "SP_multicast error(%d)", sperrno);
            return;
        }
		return_value->type = IS_LONG;
		return_value->value.lval = sperrno;
        return;
    }
}
/* }}} */

PHP_FUNCTION(spread_join) {
    zval **group, **mbox_zval;
    int *mbox;
    int sperrno, default_mbox;
    switch(ZEND_NUM_ARGS()) {
        case 2:
            if(zend_get_parameters_ex(2, &mbox_zval, &group) == FAILURE) {
                RETURN_FALSE;
            }
            break;
        default:
            WRONG_PARAM_COUNT;
            break;
    }
    default_mbox = SpG(default_conn);
    ZEND_FETCH_RESOURCE(mbox, int *, mbox_zval, default_mbox, "Spread-FD", le_conn);
    SEPARATE_ZVAL(group)
    if((*group)->type == IS_STRING) {
        char *tmpgrp;
        convert_to_string_ex(group);
        tmpgrp = estrndup((*group)->value.str.val,(*group)->value.str.len);
        if( (sperrno = SP_join(*mbox, tmpgrp)) < 0) {
            zend_error(E_WARNING, "SP_join error(%d)", sperrno);
            RETURN_FALSE;
        }
        return_value->type = IS_LONG;
        return_value->value.lval = sperrno;
        return;
    }
    else if((*group)->type == IS_ARRAY) {
        char groupnames[100][MAX_GROUP_NAME];
        HashPosition pos;
        zval *tmparr, **tmp;
        int n = 0;
        int error = 0;
        tmparr = *group;
        zend_hash_internal_pointer_reset_ex(tmparr->value.ht, &pos);
        while(zend_hash_get_current_data_ex(tmparr->value.ht, (void **) &tmp, &pos) == SUCCESS && n < 100) {
            convert_to_string_ex(tmp);
            if( (sperrno = SP_join(*mbox,  (*tmp)->value.str.val)) < 0) {
                zend_error(E_WARNING, "SP_join error(%d)", sperrno);
                error = sperrno;
            }
            n++;
            zend_hash_move_forward_ex(tmparr->value.ht, &pos);
        }
        if (error) {
            return_value->type = IS_LONG;
            return_value->value.lval = sperrno;
            return;
        } else {
            RETURN_TRUE;
        }
    }
    else {
        zend_error(E_WARNING, "SP_join reports inconsistent datatypes");

    }
}

PHP_FUNCTION(spread_receive) {
    zval **mbox_zval, **timeout_zval, *groups_zval;
    int *mbox;
    double timeout;
    struct timeval towait;
    int sperrno, default_mbox;
    int i, endmis, ret, ngrps, msize;
    int16 mtype;
    service stype;
    static int oldmsize = 0;
    static int oldgsize = 0;
    static int newmsize = (1<<15);
    static int newgsize = (1<<6);
    static char* groups=NULL;
    static char* mess=NULL;
    char sender[MAX_GROUP_NAME];
    fd_set readfs;

    switch(ZEND_NUM_ARGS()) {
        case 1:
            if(zend_get_parameters_ex(1, &mbox_zval) == FAILURE) {
                RETURN_FALSE;
            }
            towait.tv_sec = 0;
            towait.tv_usec = 0;
            break;
        case 2:
            if(zend_get_parameters_ex(2, &mbox_zval, &timeout_zval) == FAILURE) {
                RETURN_FALSE;
            }
            convert_to_double_ex(timeout_zval);
            timeout = (double)Z_DVAL_PP(timeout_zval);
            towait.tv_sec = (unsigned long)timeout;
            towait.tv_usec =
                (unsigned long)(1000000.0*(timeout-(double)towait.tv_sec));
            break;
        default:
            WRONG_PARAM_COUNT;
            break;
    }
    default_mbox = SpG(default_conn);
    ZEND_FETCH_RESOURCE(mbox, int *, mbox_zval, default_mbox, "Spread-FD", le_conn);
    FD_ZERO(&readfs);
    FD_SET(*mbox, &readfs);
    if((ret = select(*mbox+1, &readfs, NULL, &readfs, &towait))!=1) {
        RETURN_FALSE;
    }
    try_again: {
        if(oldgsize != newgsize) {
            if(groups) {
                groups = (char*) erealloc(groups, newgsize*MAX_GROUP_NAME);
            } else {
                groups = (char*) emalloc(newgsize*MAX_GROUP_NAME);
            }
            oldgsize=newgsize;
        }
        if(oldmsize != newmsize) {
            if(mess) {
                mess = (char *) erealloc(mess, newmsize);
            } else { 
                mess = (char *) emalloc(newmsize);
            }
            oldmsize = newmsize;
        }
        if((ret=SP_receive(*mbox, &stype, sender, newgsize, &ngrps, groups,
            &mtype, &endmis, newmsize, mess))<0) {
            if(ret==BUFFER_TOO_SHORT) {
                newmsize=-endmis;
                newmsize++;
                msize = oldmsize;
                goto try_again;
            }
        }
        msize = oldmsize;
    }
    /* spread does not null termintae these, so we should */
    mess[msize + 1] = '\0';
    /* we've got theanswer, let's wind up our response */
    if (array_init(return_value)==FAILURE) {
        RETURN_FALSE;
    }
    add_assoc_stringl(return_value, "message", mess, msize, 1);
    ALLOC_ZVAL(groups_zval);
    array_init(groups_zval);
    for(i = 0; i < ngrps; i++) {
        add_index_stringl(groups_zval, i, &groups[i*MAX_GROUP_NAME], strlen(&groups[i*MAX_GROUP_NAME]), 1);
    }
    add_assoc_zval(return_value, "groups", groups_zval); 
    add_assoc_long(return_value, "message_type", mtype); 
    add_assoc_stringl(return_value, "sender", sender, strlen(sender), 1); 
}
                

/* {{{ proto int spread_close([resource id])
	closes a spread connection */
PHP_FUNCTION(spread_close) {
	zval **spread_conn;
	mailbox *mbox;
	int id;
	
	switch(ZEND_NUM_ARGS()) {
		case 0:
			id = SpG(default_conn);
			break;
		case 1:
			if(zend_get_parameters_ex(1, &spread_conn) == FAILURE) {
				RETURN_FALSE;
			}
			id = -1;
			break;
		default:
			WRONG_PARAM_COUNT;
			break;
	}
	ZEND_FETCH_RESOURCE(mbox, mailbox *, spread_conn, id, "Spread-FD", le_conn);
	if (id == -1) {
		zend_list_delete(Z_RESVAL_PP(spread_conn));
	}
	if(id != -1 || (spread_conn && Z_RESVAL_PP(spread_conn)==SpG(default_conn))){
		zend_list_delete(SpG(default_conn));
		SpG(default_conn) = -1;
	}
	RETURN_TRUE;
}
/* }}} */

/* does nothing (currently) */
PHP_FUNCTION(spread_error) {
}
